/***************************************************************************
  Research Academic Computer Technology Institute,
  61 Riga Feraiou Str,26221 Patras ,Greece &
  Department of Computer Engineering and Informatics,
  University of Patras, 26500 Rion ,Patras,Greece

  Research Unit 6

  project name:Bandwidth Broker
  project team:Christos Bouras
               Dimitris Primpas
	       Kostas Stamos
	       Ioannis Pappas

  date:6 September 2005
  updated:no
  contact:bouras@cti.gr,primpas@cti.gr,stamos@cti.gr,pappasj@ceid.upatras.gr

*****************************************************************************/

/*Headers*/
#include "CentralAacrBBbase.h"
#include <iostream.h>
#include <fstream>
#include <stdlib.h>
#include <cstdlib>
extern "C" {
	#include "glpk.h"
}

/*constants*/
/*******************************************
 constants name:REQUEST,ANSWER,FIX_BWD
 		UNAVAILIABLE,DISTR_QUESTION,
		DISTR_ANSWER,CENTR_QUESTION,
		CENTR_ANSWER
mission:these constants specify the
	type of a packet
*******************************************/
#define REQUEST        0/*request packet from BBedge to BBbase for bandwidth */
#define ANSWER         1/*answer packet */
#define FIX_BWD        2/*command packet to BBedge to fix his bwd*/
#define UNAVAILABLE    3/*information packet to BBedge that BBbase is unavailiable now*/
#define DISTR_QUESTION 4 /*question packet from DistributedBBbase to BBedge if edge has bwd*/
#define DISTR_ANSWER   5 /*answer to  DistributedBBbase if BBedge has bwd*/
#define CENTR_QUESTION 6 /* question packet from CentralBBbase of what  BBedge controls */
#define CENTR_ANSWER   7 /*answer to CentralBBbase of what  BBedge controls*/
#define SIGNAL_FINISH  8 /*signals that experiment has ended, is used to empty rArray*/


#define NEGATIVE 0
#define POSITIVE 1

/**********************************************
 constants name:DOWN,ACTIVE
 mission:these constants specify the status in
         which a BBagent is
**********************************************/
#define DOWN     0
#define ACTIVE   1

/*******************************************
 constants name:PENDING,SATISFIED,
 		REJECTED,EXPIRED
mission:these constants specify the
	status in which a request is
*******************************************/
#define PENDING     0
#define SATISFIED   1
#define REJECTED    2
#define EXPIRED     3



static class CentralAacrBBbaseClass : public TclClass {
public:
	CentralAacrBBbaseClass() : TclClass("Agent/BBbase/CentralAacr") {}
	TclObject* create(int, const char*const*) {
		return (new CentralAacrBBbaseAgent());
	}
} class_CentralAacrBBbase;


/*constructor*/
CentralAacrBBbaseAgent::CentralAacrBBbaseAgent() 
{
    bind("packetSize_", &size_);
    available = TRUE;/* at the beginning,BBbase is availiable*/
    cur_dscp=2;

	Rsize = 10;		// For PBAC this is constant value, for AAC/AACR just initialization
//	bind("Rsize", &Rsize);

    bbgnt_particip=new BBagent_Participants;/* object which handles the BBAgents list*/
    max_request_buffer=0;/*FUTURE USE*/ /*if the user does not set the magnidute of the buffer, then the default is 0 */

	// Initialize statistics
	count_requests_valid_nonzero = 0;
	count_requests_invalid = 0;
	count_requests_zero = 0;
	count_accepted = 0;
	count_profit = 0;
	count_answer_time = 0;
	count_LPsize = 0;
	count_LPproblems = 0;

	// Initialize AAC/AACR parameters
	threshold = 5;
	a = 0.5;

	// Initialize AACR parameters
	resubmit_probability = 50;	// in % points
	count_resubmitted = 0;
}


/************************************************
 function name: CentralBBbaseAgent::command
 return value:int
 arguments:number of tcl command arguments,
 	   the command arguments
 mission:takes the tcl commands and transforms
         them to c++ functions & operations
*************************************************/
int CentralAacrBBbaseAgent::command(int argc, const char*const* argv)
{

	if (strcmp(argv[1], "finish") == 0) {
		// Run recv with a dummy SIGNAL_FINISH packet to make sure all requests are examined
        Packet* pkt = allocpkt();
        hdr_bbe* hdr = hdr_bbe::access(pkt);/* Access the bbe header*/
        hdr->type=SIGNAL_FINISH;  /*packet type==SIGNAL_FINISH*/
cout << "Calling recv from command for SIGNAL_FINISH\n";
		recv(pkt,0);
cout << "Finished recv called from command\n";
		return (TCL_OK);
    }

   if (argc == 3)/*if we have three(3) arguments*/
   {
      /*******************************************
       FUTURE USE
       if the tcl command is set_request_buffer,
       then the user sets the max magnitude of
       the buffer which keeps all the requests
        that cannot now be examined
      *******************************************/
      if (strcmp(argv[1], "set_request_buffer") == 0)
        {
          max_request_buffer=atoi(argv[2]);
          return (TCL_OK);
	}
   }

   return (Agent::command(argc, argv));
}


/***********************************************
 function name: CentralBBbaseAgent::recv
 return value:void
 arguments:a packet , a handler
 mission:receives and handles the packets of
         the CentralBBbaseAgent
************************************************/
void CentralAacrBBbaseAgent::recv(Packet* pkt, Handler*)
{
    /* Access the IP header for the received packet*/
    hdr_ip* hdrip = hdr_ip::access(pkt);
    /* Access the bbe header for the received packet*/
    hdr_bbe* hdr = hdr_bbe::access(pkt);

    /*******************************************
     create an instance of the request in which
     the request packet refers to
   ********************************************/
   Request pck_request(hdr->sender,hdr->endnode,hdr->bndw_request,hdr->start_time,hdr->stop_time);



    /*****************************************
     if the type of the packet is REQUEST
    ******************************************/
    if (hdr->type == REQUEST)
    {


	    /*we find the current time of the expirement*/
	    Tcl& tcl= Tcl::instance();
        tcl.evalc("set now [$ns now]");
	    char *nh = (char *)tcl.result();
        cout<<"Got a new request, at time: "<<atoi(nh);

	   /*********************************
	   IF the pck_request is still valid
	   **********************************/
	   if(atoi(nh) < pck_request.GetStart_time() )
	   {
	    cout<<" o.k. the time of the request is valid"<<"\n";

		// If the packet requests 0 bandwidth, just accept it
		if (pck_request.GetBndw_request() == 0) {

	   		cout << " Valid request, but no bwd requested\n";
			count_requests_zero++;

			/****************************
			send positive answer to
			the request sender
			*****************************/
			dst_.addr_ = hdr->sender;/*set the destination address of the packet*/
			/****************************
         	 allocate & send a new packet
	        *****************************/
			Packet* next_pkt = allocpkt();
			hdr_ip* ippkt = hdr_ip::access(next_pkt);
			ippkt->DSCP_=1;
			hdr_bbb* next_hdr      = hdr_bbb::access(next_pkt);
			next_hdr->answer       = POSITIVE;
			next_hdr->type         = ANSWER;
			next_hdr->sender       = hdr->sender;
			next_hdr->endnode      = hdr->endnode;
			next_hdr->dscp         = cur_dscp;
			next_hdr->bndw_request =hdr->bndw_request;
			next_hdr->start_time   =hdr->start_time;
			next_hdr->stop_time    =hdr->stop_time;
			cur_dscp=cur_dscp+POSITIVE;
			send(next_pkt,0);

			Request newrequest(hdr->sender,hdr->endnode,hdr->bndw_request,hdr->start_time,hdr->stop_time);
			newrequest.SetStatus(SATISFIED);
			newrequest.SetArrival_time(atoi(nh));
			newrequest.SetAnswer_time(atoi(nh));
			list_request->add_to_list(newrequest);/*create a new record to request database*/
		} // end of valid-but-zero-requesting packet handling
		else {
cout << " Valid non-zero new request, requests bwd:" << hdr->bndw_request;
cout << " \tDuration " << hdr->start_time << "-" << hdr->stop_time;
cout << " \t\tDirection " << hdr->sender << "->" << hdr->endnode << endl;
			count_requests_valid_nonzero++;

			/******************************************
			create a new record to request database
			accomplishes the request according to his
			database
            ******************************************/

			dst_.addr_ = hdr->sender;/*set the destination address of the packet*/

			Request newrequest(hdr->sender,hdr->endnode,hdr->bndw_request,hdr->start_time,hdr->stop_time);
			newrequest.SetArrival_time(atoi(nh));
			list_request->add_to_list(newrequest);/*create a new record to request database*/

			// If BB is unavailable, put request in Wq
			if (available) {
				rArray.push_back(newrequest);
			}
			else {
				wqArray.add_to_list(newrequest);
			}

		}
	   }/*end of if the request is valid*/

	   /*************************************
	    ELSE Negative answer to the sender
	   /*************************************/
	   else	// invalid packet
	   {
	   	cout<<" Sorry the time of the request is invalid"<<"\n";
		count_requests_invalid++;

		/****************************
		 send negative answer to
		 the request sender
		*****************************/
		dst_.addr_ = hdr->sender;/*set the destination address of the packet*/
		/****************************
         	 allocate & send a new packet
	        *****************************/
		Packet* next_pkt = allocpkt();
		hdr_ip* ippkt = hdr_ip::access(next_pkt);
		ippkt->DSCP_=1;
		hdr_bbb* next_hdr      = hdr_bbb::access(next_pkt);
		next_hdr->answer       = NEGATIVE;
		next_hdr->type         = ANSWER;
		next_hdr->sender       = hdr->sender;
		next_hdr->endnode      = hdr->endnode;
		next_hdr->dscp         = cur_dscp;
		next_hdr->bndw_request =hdr->bndw_request;
		next_hdr->start_time   =hdr->start_time;
		next_hdr->stop_time    =hdr->stop_time;
		cur_dscp=cur_dscp+0;
		send(next_pkt,0);


		Request newrequest(hdr->sender,hdr->endnode,hdr->bndw_request,hdr->start_time,hdr->stop_time);
		newrequest.SetStatus(REJECTED);
		list_request->add_to_list(newrequest);/*create a new record to request database*/
       } // end of invalid packet handling


	}/*end of first part of packet handling for packet_type==REQUEST*/

	// This will be executed for both REQUEST and the dummy SIGNAL_FINISH requests
	if ((hdr->type == REQUEST) || (hdr->type == SIGNAL_FINISH)) {

		// If R set has filled, examine requests in it and then empty it
		// Also, if we are at the SIGNAL_FINISH case, we examine requests until rArray is empty
		while ( (rArray.size() >= Rsize) || 
				( (hdr->type == SIGNAL_FINISH) && (!rArray.empty()) )
			  ) {

			count_LPsize += rArray.size();
			count_LPproblems++;

cout << "bbgnt_particip size:" << bbgnt_particip->GetSize_of_list() << endl;
			// Becomes unavailable while computing requests
			available = FALSE;

			// Since the algorithm is working with the "VPN hose model", all requests are competing
			// for a single (virtual) link. So simply get one request in order to get the handlers
			// (BBAgent etc.) for the specific link (which should be the same for all requests because
			// of the "hose" model)
			Tcl& tcl= Tcl::instance();
			tcl.evalc("$ns compute-routes");	// Compute all next hops for the topology
			tcl.evalc("set rlg [$ns get-routelogic]");	// Return a handle to the route loggic object (rlg)

			Request req = rArray[0];	// Just get the first request, it doesn't matter anyway
cout << "For getting BBagents, we use request " << req.GetSender() << "->" << req.GetEndnode() << " wants " << req.GetBndw_request() << " from " << req.GetStart_time() << "-" << req.GetStop_time() << endl;

			// Return the id of the next hop from sender->endnode
			tcl.evalf("set temp [$rlg lookup %d %d]",(int)req.GetSender(),(int)req.GetEndnode());
			char *nh = (char *)tcl.result();
			int sender_id = req.GetSender();
			int nexthop_id = atoi(nh);
			BBagent bb_agent_sender(sender_id);
cout << "Created BBagent with id (sender_id):" << sender_id << endl;
			BBagent bb_agent_nexthop(nexthop_id);
cout << "Created BBagent with id (nexthop_id):" << nexthop_id << endl;

			// Setup LP problem
			// LPX_UP: upper bound, LPX_LO: lower bound, LPX_DB: double bounded
			LPX *lp;
			int ia[1+1000], ja[1+1000];
			double ar[1+1000], Z;
			lp = lpx_create_prob();		// Create problem object
			lpx_set_prob_name(lp, "AAC solving");	// Name
			lpx_set_class(lp, LPX_MIP);		// Mixed integer problem
			lpx_set_obj_dir(lp, LPX_MAX);	// Maximization
			
			double earliest_tstart = 1000000000;	// Just a huge value
			double latest_tend = 0;
			double total_duration = 0;
			int rArray_size = rArray.size();

			// Fill the objective function parameters array ("columns")
			// (sum of bandwidth*duration for each request)
			double* obj = new double[rArray_size];
			lpx_add_cols(lp, rArray_size);	// Number of columns
			cout << "Number of columns (rArray_size): " << rArray_size << endl;
			for (int i_c=0; i_c<rArray_size; i_c++) {
				Request req1 = rArray[i_c];

				// Find the earliest Tstart
				if (req1.GetStart_time()<earliest_tstart || i_c==0) {
					earliest_tstart = req1.GetStart_time();
				}
				// Find the latest Tend
				if (req1.GetStop_time()>latest_tend || i_c==0) {
					latest_tend = req1.GetStop_time();
				}

				// lpx functions count columns starting from index 1, not 0
				char colname[10];
				sprintf(colname,"col_x%d",i_c+1);
				lpx_set_col_name(lp, i_c+1, colname);
				lpx_set_col_bnds(lp, i_c+1, LPX_DB, 0.0, 1.0);	// Set as binary
				lpx_set_obj_coef(lp, i_c+1, req1.GetBndw_request()*(req1.GetStop_time()-req1.GetStart_time()) );
				lpx_set_col_kind(lp, i_c+1, LPX_IV);	// Set the column as integer value
			}
			total_duration = latest_tend - earliest_tstart;
			if (total_duration <=0) {
				cout << "Total duration for LP is zero or below! Check input file for invalid requests";
			}

			// Fill the constraints values array (the available bandwidth at each second)
			// Also fill the direction of each unequation for the constraints
			lpx_add_rows(lp, (int)total_duration);	// Number of rows
			cout << "Number of rows (total_duration): " << total_duration << endl;
			for (int i_r=0; i_r<total_duration; i_r++) {
				int avl_bwd = bbgnt_particip->agent_query_bwd_available(bb_agent_sender, nexthop_id, (int)(i_r+earliest_tstart), 1);
				if (avl_bwd == -1) {	// Error
					cout << "Error retrieving available bandwidth! (Called query with backup option)" << endl;
					return;
				}
cout << "Available bwd for time " << i_r+earliest_tstart << " is:" << avl_bwd << endl;
				
				// lpx functions count columns starting from index 1, not 0
				char rowname[10];
				sprintf(rowname,"row%d",i_r+1);
				lpx_set_row_name(lp, i_r+1, rowname);
				lpx_set_row_bnds(lp, i_r+1, LPX_UP, 0.0, avl_bwd);
			}

			// Fill ia, ja, ar arrays
			// ia stores the row indices, ja the column indices, ar the values
			int index = 1;
			for (int col_i=0; col_i<rArray_size; col_i++) {
//cout << "col_i=" << col_i << endl;
				Request req2 = rArray[col_i];
				for (int row_i=0; row_i<total_duration; row_i++) {
//cout << "row_i=" << row_i << endl;
//cout << "Filling ia[" << index << "]=" << row_i+1 <<", ja[" << index << "]=" << col_i+1 <<", ar[" << index << "]=" << req2.GetBndw_request() << "...";
					if ( (row_i >= req2.GetStart_time()-earliest_tstart) &&
						 (row_i < req2.GetStop_time()-earliest_tstart) ) {
						ia[index] = row_i+1;	// Counting in ia starts from 1, not 0
						ja[index] = col_i+1;	// Counting in ja starts from 1, not 0
						ar[index] = req2.GetBndw_request();	// Counting in ar starts from 1, not 0
						index++;
//cout << "filled." << endl;
					}
					else {
						// No filling with 0 necessary (ar[index] = 0) since only non-zero
						// elements are filled. So we need not do anything in the "else" part,
						// and the index of the ar array does not advance
//cout << "NOT filled." << endl;
					}
				}
			}
			index--;	// Reverse the final increment of index

			lpx_print_prob(lp, "lpx_problem");
			
			lpx_load_matrix(lp, index, ia, ja, ar);
			lpx_simplex(lp);
			//Z = lpx_mip_obj_val(lp);
			Z = lpx_get_obj_val(lp);

			//lpx_print_mip(lp, "lpx_problem_solution");
			lpx_print_prob(lp, "lpx_problem_solution");

			int result = 0;
			for (int k=0; k<rArray_size; k++) {
				//result = lpx_mip_col_val(lp, k+1);	// Counting in ia starts from 1, not 0
				result = lpx_get_col_prim(lp, k+1);	// Counting in ia starts from 1, not 0
				cout << "Request " << k+1 << " (in current Rset) is accepted? " << result << endl;	// Print whether accepted or not

				Tcl& tcl= Tcl::instance();
				tcl.evalc("set now [$ns now]");
				char *nh = (char *)tcl.result();

				Request req3 = rArray[k];
				req3.SetAnswer_time(atoi(nh));
cout << " Request stats: requests bwd:" << req3.GetBndw_request();
cout << " \tDuration " << req3.GetStart_time() << "-" << req3.GetStop_time();
cout << " \t\tDirection " << req3.GetSender() << "->" << req3.GetEndnode();
cout<<" \t\tAnswer Time: "<< atoi(nh) << endl;

				count_answer_time += req3.GetAnswer_time() - req3.GetArrival_time();

				// Update the bandwidth information (both directions) for the accepted requests
				if (result == 1) {

					count_accepted++;
					count_profit += ( req3.GetStop_time() - req3.GetStart_time() ) * req3.GetBndw_request();

					bbgnt_particip->agent_remove_bwd(bb_agent_sender, nexthop_id, req3.GetBndw_request(), req3.GetStart_time(), req3.GetStop_time());
					bbgnt_particip->agent_remove_bwd(bb_agent_nexthop, sender_id, req3.GetBndw_request(), req3.GetStart_time(), req3.GetStop_time());

					/****************************
					send positive answer to
					the request sender
					*****************************/
					dst_.addr_ = req3.GetSender();/*set the destination address of the packet*/
					/****************************
		         	 allocate & send a new packet
				    *****************************/
					Packet* next_pkt = allocpkt();
					hdr_ip* ippkt = hdr_ip::access(next_pkt);
					ippkt->DSCP_=1;
					hdr_bbb* next_hdr      = hdr_bbb::access(next_pkt);
					next_hdr->answer       = POSITIVE;
					next_hdr->type         = ANSWER;
					next_hdr->sender       = req3.GetSender();
					next_hdr->endnode      = req3.GetEndnode();
					next_hdr->dscp         = cur_dscp;
					next_hdr->bndw_request = req3.GetBndw_request();
					next_hdr->start_time   = req3.GetStart_time();
					next_hdr->stop_time    = req3.GetStop_time();
					cur_dscp=cur_dscp+POSITIVE;
					send(next_pkt,0);

				}
				else {
cout << "\t\t\t\tRejected request...";
					// For AACR, we make the assumption/simulation that each rejected request, with
					// a specific probability is re-introduced to the R set, and therefore re-examined
					// at the next LP instance solution
					
					// Probability calculation:
					// Produce a random integer in the range 0...100
					// If below the resubmit_probability parameter, the request is resubmitted 
					if (resubmit_probability >= (rand()%101)) {
cout << " resubmiteed" << endl;
						resubmittedArray.add_to_list(req3);
						count_resubmitted++;
					}
				}
			}

			lpx_delete_prob(lp);	// Free memory for LPX object

			int comp_time = rArray.size();	// A convention, that comp time is analogous to LP problem size
			// For AAC algorithm, here we have to Re-calculate Rsize
			if ((comp_time-threshold)>=threshold) {
				// comp_time twice or more than the threshold, decrease to minimum
				Rsize = 1;
			}
			else if ((comp_time-threshold)>0) {
				// comp_time more than the threshold (but less than two-fold), decrease
				// by a factor (0..1) analogous to how much more it exceeds the threshold
				Rsize = (1-(comp_time-threshold)/threshold)*Rsize;
			}
			else {
				// comp_time equal or less than the threshold, increase by a factor
				Rsize = Rsize + (wqArray.GetSize_of_list()-Rsize)*a;
			}
			// Check that rSize does not get invalid values
			if (Rsize <= 0) {
				Rsize = 1;
			}


			// Empty R set
			rArray.clear();

			// If Wq contains requests, start transfering them to R
			while ((wqArray.GetSize_of_list() > 0) && (rArray.size() < Rsize)) {
				Request req4 = wqArray.get_first_rqst();
				wqArray.remove_from_list(req4);
				
				// Time-shift resubmitted requests
				req4.SetStart_time(req4.GetStart_time()+10);
				req4.SetStop_time(req4.GetStop_time()+10);

				rArray.push_back(req4);
			}

			// Become available again
			available = true;

		}	// end of examining R set requests

		// AACR algorithm-specific
		// If we have resubmitted requests, transfer them to Wq
		while (resubmittedArray.GetSize_of_list() > 0) {
			Request req5 = resubmittedArray.get_first_rqst();
			resubmittedArray.remove_from_list(req5);
			wqArray.add_to_list(req5);
		}

		if (hdr->type == SIGNAL_FINISH) {	// That's it, exp is finished, print stats
			cout << endl;
			cout << "\t\t TOTAL STATS" << endl;
			cout << "\t\t" << count_requests_valid_nonzero << "\t valid requests" << endl;
			cout << "\t\t" << count_requests_zero << "\t requests for 0 bwd" << endl;
			cout << "\t\t" << count_requests_invalid << "\t invalid requests" << endl;
			cout << "\t\t" << count_accepted << "\t accepted requests" << endl;
			cout << "\t\t" << count_profit << "\t total profit" << endl;
			cout << "\t\t" << ((double)count_answer_time)/(count_requests_valid_nonzero + count_requests_zero) << "\t average time waiting for answer (total answer time=" << count_answer_time << ")" << endl;
			cout << "\t\t" << ((double)count_LPsize)/count_LPproblems << "\t average size of LP problem (total LP size=" << count_LPsize << ")"  << endl;
			cout << "\t\t" << count_resubmitted << "\t resubmitted requests" << endl;
			cout << endl;
		}

    }/*end of packet handling ,packet_type==REQUEST or SIGNAL_FINISH*/



    /******************************************
     if the type of the packet is CENTR_ANSWER
    *******************************************/
    /******************************************
     this packet is an inform packet
      for the condition of a node.
     A BBedge agent send it  to the
     CentralBBbase either because BBedge is now
     up or because BBedge Agent answers to a
     question of the CentralBBbase
    *******************************************/
    /******************************************
     answer from a node. enter the
     information to the list(database)
    ******************************************/
    else if(hdr->type == CENTR_ANSWER)
    {
       PacketData* data_;
        data_=pkt->userneighdata();/*pass the data of the packet into a AppData instance*/

       /*rebuilding the list*/
       list<neighbour> nnghbr_list;

       int i=0;
       int j=0;
       int size=pkt->dataneighlen();/*data magnitude*/
       int* data;
       data=data_->neighdata();
       /*cout<<"the magnitude of data is "<< size <<"\n"; */
       int nsize=size/2;
       for(j=0;j!=nsize;j++)
       {
            int x;
            int y;

            x=data[i];
	    cout<<"n_id is "<< x <<"\n";
	    i=i+1;
            y=data[i];
	    cout<<"bwdth is "<< y <<"\n";
	    i=i+1;
            neighbour neighb(x,y,data_->start_time(),data_->stop_time());
            nnghbr_list.push_back(neighb);

       }

       /*create an object of BBagent in order to make update of the list, the particular BBagent*/
       BBagent n_agent(hdrip->saddr(),nnghbr_list,hdr->status);/*function saddr() does this:*/
	    				    /*nsaddr_t& saddr() { return (src_.addr_); }*/
	                                    /*returning the address of the sender of the packet*/


	/*add the new information about BBagent into the list , with the function update list*/
       bbgnt_particip->update_list(n_agent);
	   cout<<"end of packet_type CENTR_ANSWER, created BBagent with id:" << hdrip->saddr() << endl;

    }/*end of packet handling ,packet_type==CENTR_ANSWER*/

    Packet::free(pkt);
}/*end of function "recv"*/
