/***************************************************************************
  Research Academic Computer Technology Institute,
  61 Riga Feraiou Str,26221 Patras ,Greece &
  Department of Computer Engineering and Informatics,
  University of Patras, 26500 Rion ,Patras,Greece

  Research Unit 6

  project name:Bandwidth Broker
  project team:Christos Bouras
               Dimitris Primpas
	       Kostas Stamos
	       Ioannis Pappas

  date:20 January 2005
  updated:no
  contact:bouras@cti.gr,primpas@cti.gr,stamos@cti.gr,pappasj@ceid.upatras.gr

*****************************************************************************/

/*Headers */
#include "BBbase.h"
#include <iostream.h>

/*constants*/
/*******************************************
 constants name:REQUEST,ANSWER,FIX_BWD
 		UNAVAILIABLE,DISTR_QUESTION,
		DISTR_ANSWER,CENTR_QUESTION,
		CENTR_ANSWER
mission:these constants specify the
	type of a packet
*******************************************/
#define REQUEST        0/*request packet from BBedge to BBbase for bandwidth */
#define ANSWER         1/*answer packet */
#define FIX_BWD        2/*command packet to BBedge to fix his bwd*/
#define UNAVAILABLE    3/*information packet to BBedge that BBbase is unavailiable now*/
#define DISTR_QUESTION 4 /*question packet from DistributedBBbase to BBedge if edge has bwd*/
#define DISTR_ANSWER   5 /*answer to  DistributedBBbase if BBedge has bwd*/
#define CENTR_QUESTION 6 /* question packet from CentralBBbase of what  BBedge controls */
#define CENTR_ANSWER   7 /*answer to CentralBBbase of what  BBedge controls*/


#define NEGATIVE 0
#define POSITIVE 1

/**********************************************
 constants name:DOWN,ACTIVE
 mission:these constants specify the status in
         which a BBagent is
**********************************************/
#define DOWN     0
#define ACTIVE   1

/*******************************************
 constants name:PENDING,SATISFIED,
 		REJECTED,EXPIRED
mission:these constants specify the
	status in which a request is
*******************************************/
#define PENDING     0
#define SATISFIED   1
#define REJECTED    2
#define EXPIRED     3



int hdr_bbe::offset_;

static class BBbaseHeaderClass : public PacketHeaderClass {
public:
	BBbaseHeaderClass() : PacketHeaderClass("PacketHeader/BBbase",
					      sizeof(hdr_bbe)) {
		bind_offset(&hdr_bbe::offset_);
	}
} class_bbehdr;

static class BBbaseClass : public TclClass {
public:
	BBbaseClass() : TclClass("Agent/BBbase") {}
	TclObject* create(int, const char*const*) {
		return (new BBbaseAgent());
	}
} class_BBbase;


/* constructor */
BBbaseAgent::BBbaseAgent() : Agent(PT_BBE)
{
   bind("packetSize_", &size_);
   available = TRUE;/* at the beginning,BBbase is availiable*/
   cur_dscp=2;
   list_request= new List_Requests; /* object which handles the request list*/
   list_request_buffer= new List_Requests; /* object which handles the request buffer list*/
   max_request_buffer=0;/*if the user does not set the magnidute of the buffer, then the default is 0 */
}


/************************************************
 function name: BBbaseAgent::command
 return value:int
 arguments:number of tcl command arguments,
 	   the command arguments
 mission:takes the tcl commands and transforms
         them to c++ functions & operations
*************************************************/
int BBbaseAgent::command(int argc, const char*const* argv)
{
    if (argc == 3)/*if we have three(3) arguments*/
    {
      /*******************************************
       if the tcl command is set_request_buffer,
       then the user sets the max magnitude of
       the buffer which keeps all the requests
        that cannot now be examined
      *******************************************/
      if (strcmp(argv[1], "set_request_buffer") == 0)
      {
          max_request_buffer=atoi(argv[2]);
          return (TCL_OK);
      }
    }
    if (argc == 4)/*if we have four(4) arguments*/
    {
      if (strcmp(argv[1], "sendto") == 0)
        {
         return (sendto((nsaddr_t)atoi(argv[2]),atoi(argv[3])));
	}
    }
    return (Agent::command(argc, argv));
}


/***********************************************
 function name: BBbaseAgent::sendto
 return value:int
 arguments:a node address,bandwidth
 mission:none
************************************************/
int BBbaseAgent::sendto(nsaddr_t dstid, int bwd)
{
    return (TCL_OK);
}


/***********************************************
 function name: BBbaseAgent::recv
 return value:void
 arguments:a packet , a handler
 mission:receives and handles the packets of
         the BBbaseAgent
************************************************/
void BBbaseAgent::recv(Packet* pkt, Handler*)
{
    /* Access the IP header for the received packet*/
    hdr_ip* hdrip = hdr_ip::access(pkt);
    /* Access the bbe header for the received packet*/
    hdr_bbe* hdr = hdr_bbe::access(pkt);


    /*****************************************
     if the type of the packet is REQUEST
    ******************************************/
    if (hdr->type == REQUEST)
    {

      /*******************************************
       create an instance of the request in which
       the request packet refers to
      ********************************************/
      Request pck_request(hdr->sender,hdr->endnode,hdr->bndw_request,hdr->start_time,hdr->stop_time);

        /******************************************
	 IF this is the first bwd request packet a
	 BBedge sends(hop_num==1) but the BBbase
	 is unavailiable ,
	  IF the list_request_buffer is not full
	   add this request to the list
	  ELSE send back to the BBedge that the
	   Agent is UNAVAILIABLE
	*******************************************/
	if (hdr->hop_num==1 && available==FALSE)
        {
	    if(list_request_buffer->GetSize_of_list() < max_request_buffer)
	    {
	    	/*cout<<"hello from hdr->hop_num==1 && available==FALSE IF"<<"\n";*/
	      /*create a new record to request_buffer database*/
              list_request_buffer->add_to_list(pck_request);
	      /*cout<<"now the list has magnitude "<<list_request_buffer->GetSize_of_list()<<"\n";*/
	    }
	    else
            {
             dst_.addr_ = hdr->sender;/*set the destination address of the packet*/
	     /****************************
	     allocate & send a new packet
	     *****************************/
             Packet* next_pkt = allocpkt();
             hdr_ip* ippkt = hdr_ip::access(next_pkt);
             ippkt->DSCP_=1;
             hdr_bbb* next_hdr = hdr_bbb::access(next_pkt);
             next_hdr->type = UNAVAILABLE;
             next_hdr->endnode = hdr->endnode;
             send(next_pkt,0);
	    }

	}
	/*******************************************
	 ELSE
	********************************************/
        else
        {
	    /******************************************
             IF this is the first bwd request packet a
	     BBedge sends(hop_num==1) and BBbase is
	     availiable
	    *******************************************/
	    if(hdr->hop_num==1 && available==TRUE)
	    {

	     /*************************************
	      IF the list_request_buffer is empty
	      examine the request
	     **************************************/
	     if(list_request_buffer->GetSize_of_list() == 0)
	     {
	       /****************************************
                BBbase becomes unavailiable
	        create a new record to request database
	        send to the request sender  a
	        question(request) for bandwidth
	       ****************************************/
               dst_.addr_ = hdr->sender; /*set the destination address of the packet*/
	                                 /*the packet will be sent first to request sender */
	       				 /* to ask him for bwd*/
               available=FALSE; /*BBbase unavailiable*/
	       Request newrequest(hdr->sender,hdr->endnode,hdr->bndw_request,hdr->start_time,hdr->stop_time);
	       list_request->add_to_list(newrequest);/*create a new record to request database*/

   	       /****************************
	        allocate & send a new packet
	       *****************************/
               Packet* next_pkt = allocpkt();
               hdr_ip* ippkt = hdr_ip::access(next_pkt);
               ippkt->DSCP_=1;
               hdr_bbb* next_hdr = hdr_bbb::access(next_pkt);
               next_hdr->sender = hdr->sender;
               next_hdr->endnode = hdr->endnode;
               next_hdr->type = REQUEST;
               next_hdr->bndw_request = hdr->bndw_request;
               next_hdr->hop_num = hdr->hop_num;
               next_hdr->start_time = hdr->start_time;
	       next_hdr->stop_time = hdr->stop_time;
               send(next_pkt,0);

	     }/*end of " if (the list_request_buffer is empty)" */


  	     /************************************
	      IF the pck_request is the first in
	      the list_request_buffer
	      remove it from the list
              check
	      IF the pck_request is still valid
	       examine the request
	      ELSE Negative answer to the sender
	     ************************************/
	     else if( pck_request==list_request_buffer->get_first_rqst() )
	     {
	       /*cout<<"hello from else if( pck_request==list_request_buffer->get_first_rqst() )" << "\n";*/
               list_request_buffer->remove_from_list(pck_request);

	       /*we find the current time of the expirement*/
	       Tcl& tcl= Tcl::instance();
               tcl.evalc("set now [$ns now]");
	       char *nh = (char *)tcl.result();
               cout<<"the time now is "<<atoi(nh)<<"\n";

    	       /*********************************
	        IF the pck_request is still valid
	       **********************************/
	       if(atoi(nh) < pck_request.GetStart_time() )
	       {
	        cout<<"o.k. the time of the request is valid"<<"\n";

	        /****************************************
                 BBbase becomes unavailiable
	         create a new record to request database
	         send to the request sender  a
	         question(request) for bandwidth
	        ****************************************/
                dst_.addr_ = hdr->sender; /*set the destination address of the packet*/
	                                 /*the packet will be sent first to request sender */
	       				 /* to ask him for bwd*/
                available=FALSE; /*BBbase unavailiable*/
	        Request newrequest(hdr->sender,hdr->endnode,hdr->bndw_request,hdr->start_time,hdr->stop_time);
	        list_request->add_to_list(newrequest);/*create a new record to request database*/

   	        /****************************
	         allocate & send a new packet
	        *****************************/
                Packet* next_pkt = allocpkt();
                hdr_ip* ippkt = hdr_ip::access(next_pkt);
                ippkt->DSCP_=1;
                hdr_bbb* next_hdr = hdr_bbb::access(next_pkt);
                next_hdr->sender = hdr->sender;
                next_hdr->endnode = hdr->endnode;
                next_hdr->type = REQUEST;
                next_hdr->bndw_request = hdr->bndw_request;
                next_hdr->hop_num = hdr->hop_num;
                next_hdr->start_time = hdr->start_time;
	        next_hdr->stop_time = hdr->stop_time;
                send(next_pkt,0);

               }/*end of if the request is valid*/

	       /*************************************
	        ELSE Negative answer to the sender
	       /*************************************/
	       else
	       {
	   	cout<<"sorry the time of the request is invalid"<<"\n";
	        /****************************
		 send negative answer to
		 the request sender
		*****************************/
		dst_.addr_ = hdr->sender;/*set the destination address of the packet*/
		/****************************
         	 allocate & send a new packet
	        *****************************/
		Packet* next_pkt = allocpkt();
		hdr_ip* ippkt = hdr_ip::access(next_pkt);
		ippkt->DSCP_=1;
		hdr_bbb* next_hdr      = hdr_bbb::access(next_pkt);
		next_hdr->answer       = NEGATIVE;
		next_hdr->type         = ANSWER;
		next_hdr->sender       = hdr->sender;
		next_hdr->endnode      = hdr->endnode;
		next_hdr->dscp         = cur_dscp;
		next_hdr->bndw_request =hdr->bndw_request;
		next_hdr->start_time   =hdr->start_time;
		next_hdr->stop_time    =hdr->stop_time;
		cur_dscp=cur_dscp+0;
		send(next_pkt,0);


		Request newrequest(hdr->sender,hdr->endnode,hdr->bndw_request,hdr->start_time,hdr->stop_time);
		newrequest.SetStatus(REJECTED);
		list_request->add_to_list(newrequest);/*create a new record to request database*/

		/**************************************************
		 IF the list_request_buffer is now empty do nothing
                 ELSE call recv with argument the packet which holds
		 the first request of the list_request_buffer
		****************************************************/
		if(list_request_buffer->GetSize_of_list()==0)
		{}
		else
		{
		  Request demand=list_request_buffer->get_first_rqst();

		  /*cout<<"Hello "<<"\n";*/
		  /*demand.ShowRequest();*/
		  /*cout<<"Hello  again"<<"\n";*/

                  Packet* pkt = allocpkt();
                  hdr_ip* hdrip = hdr_ip::access(pkt);/* Access the IP header*/
                  hdrip->DSCP_=1;
                  hdr_bbe* hdr = hdr_bbe::access(pkt);/* Access the bbe header*/
                  hdr->sender =demand.GetSender();  /*address of the node(sender) who sends the request for bandwidth */
                  hdr->endnode =demand.GetEndnode(); /*address of the last node of the request */
                  hdr->hop_num=1;
                  hdr->type=REQUEST;  /*packet type==REQUEST*/
                  hdr->bndw_request=demand.GetBndw_request();
                  hdr->start_time=demand.GetStart_time();/*when the allocation of the bandwith begins */
                  hdr->stop_time=demand.GetStop_time();/*when the allocation of the bandwith ends */
                  recv(pkt, 0) ;
		}


	       }/*end of "else(the request that was in the list_request_buffer is invalid cause*/
	        /* the time he requested, has already passed) " */




	     }/*end of "else if( pck_request==list_request_buffer->get_first_rqst() )" */


	     /****************************************************
	      ELSE the list is not empty && no request from the
	      list_request_buffer so
	       IF the list is not full add to list_request_buffer
	       ELSE send back to the BBedge that the
	        Agent is UNAVAILIABLE
   	     ****************************************************/
             else
	     {
	       /*cout<<"i am at else where agent is availiable but the list is not empty"<<"\n";*/

	       if(list_request_buffer->GetSize_of_list() < max_request_buffer)
	       {
	         /*create a new record to request_buffer database*/
                 list_request_buffer->add_to_list(pck_request);

		 /*cout<<"Hello another"<<"\n";*/

	       }

	       else
               {
                dst_.addr_ = hdr->sender;/*set the destination address of the packet*/
	        /****************************
	        allocate & send a new packet
	        *****************************/
                Packet* next_pkt = allocpkt();
                hdr_ip* ippkt = hdr_ip::access(next_pkt);
                ippkt->DSCP_=1;
                hdr_bbb* next_hdr = hdr_bbb::access(next_pkt);
                next_hdr->type = UNAVAILABLE;
                next_hdr->endnode = hdr->endnode;
                send(next_pkt,0);
	       }/*end of "else (no request from list_request_buffer and no empty list_request_buffer )  " */


	     }/*end of the third case in which the BBbase will fall if he is availiable*/

	    }/*end of else(case BBbase is availiable)*/



	    /********************************************
	     ELSE IF the BBbase is unavailiable,
	     meaning that we have a packet like
	      hop==2 or hop==3,
	      we find the first node next to him
	      (node from where the packet came from)
	      which is on the path of the bwd request
	      and we send him a REQUEST packet
	    *********************************************/
	    else if(available==FALSE)
	    {
               Tcl& tcl= Tcl::instance();
               tcl.evalc("$ns compute-routes");
               tcl.evalc("set rlg [$ns get-routelogic]");
               tcl.evalf("set temp [$rlg lookup %d %d]",(int)hdr->sender,(int)hdr->endnode);
               char *nh = (char *)tcl.result();
               for(int i=2;i!=hdr->hop_num;i++)
               {
                 tcl.evalf("set temp [$rlg lookup %d %d]",(int)atoi(nh),(int)hdr->endnode);
                 nh = (char *)tcl.result();
               }

               dst_.addr_ = atoi(nh);/*set the destination address of the packet*/
	       			     /*we found him and we set him as the address*/
	       			     /*of the next sent packet*/

	       /****************************
	       allocate & send a new packet
	       *****************************/
               Packet* next_pkt = allocpkt();
               hdr_ip* ippkt = hdr_ip::access(next_pkt);
               ippkt->DSCP_=1;
               hdr_bbb* next_hdr = hdr_bbb::access(next_pkt);
               next_hdr->sender = hdr->sender;
               next_hdr->endnode = hdr->endnode;
               next_hdr->type = REQUEST;
               next_hdr->bndw_request = hdr->bndw_request;
               next_hdr->hop_num = hdr->hop_num;
               next_hdr->start_time = hdr->start_time;
	       next_hdr->stop_time = hdr->stop_time;
               send(next_pkt,0);
	    }/*end of else if(case BBbase is unavailiable)*/


        }/* end of ELSE */

    }/*end of packet handling, packet_type==REQUEST*/


    /*****************************************
     if the type of the packet is ANSWER
    ******************************************/
    else
    {
        /*************************************
	 first send the answer to the sender
	**************************************/
	dst_.addr_ = hdr->sender; /*set the destination address of the packet*/
	/****************************
	 allocate & send a new packet
	*****************************/
        Packet* next_pkt = allocpkt();
        hdr_ip* ippkt = hdr_ip::access(next_pkt);
        ippkt->DSCP_=1;
        hdr_bbb* next_hdr      = hdr_bbb::access(next_pkt);
        next_hdr->answer       = hdr->answer;
        next_hdr->type         = ANSWER;
        next_hdr->sender       = hdr->sender;
        next_hdr->endnode      = hdr->endnode;
        next_hdr->start_time   = hdr->start_time;
        next_hdr->stop_time    = hdr->stop_time;
        next_hdr->dscp         = cur_dscp;/* the dscp of the application if answer is positive*/
        next_hdr->bndw_request =hdr->bndw_request;
        cur_dscp=cur_dscp+hdr->answer; /*if the answer is positive, the dscp that */
				       /* BBbase gives to the applications  */
				       /* increases by one ,else stays the same*/
        send(next_pkt,0);

        /***********************************
	 if the answer is positive,BBbase
	 Agent sends command packets to
	 every BBedge Agent of the path
	 of request to fix his bwd
	************************************/
        if (hdr->answer==POSITIVE)
        {
            Tcl& tcl= Tcl::instance();
            tcl.evalc("$ns compute-routes");
            tcl.evalc("set rlg [$ns get-routelogic]");
            tcl.evalf("set temp [$rlg lookup %d %d]",(int)hdr->sender,(int)hdr->endnode);
            char *nh = (char *)tcl.result();
            int hop1 = hdr->sender;
            int hop2 = atoi(nh);

            while(1)
            {
                dst_.addr_=hop1;/*set the destination address of the packet*/
	        /****************************
	        allocate & send a new packet
	        *****************************/
                Packet* next_pkt1 = allocpkt();
                hdr_ip* ippkt1 = hdr_ip::access(next_pkt1);
                ippkt1->DSCP_=1;
                hdr_bbb* next_hdr1 = hdr_bbb::access(next_pkt1);
                next_hdr1->type = FIX_BWD;
                next_hdr1->bndw_request = hdr->bndw_request;
                next_hdr1->endnode = hop2;
	        next_hdr1->start_time = hdr->start_time;
                next_hdr1->stop_time = hdr->stop_time;
                send(next_pkt1,0);


                dst_.addr_=hop2;/*set the destination address of the packet*/
		/****************************
	         allocate & send a new packet
	        *****************************/
                Packet* next_pkt2 = allocpkt();
                hdr_ip* ippkt2 = hdr_ip::access(next_pkt2);
                ippkt2->DSCP_=1;
                hdr_bbb* next_hdr2 = hdr_bbb::access(next_pkt2);
                next_hdr2->type = FIX_BWD;
                next_hdr2->bndw_request = hdr->bndw_request;
                next_hdr2->endnode = hop1;
	        next_hdr2->start_time = hdr->start_time;
                next_hdr2->stop_time = hdr->stop_time;
                send(next_pkt2,0);

                if (hop2 == hdr->endnode)
                    break;

                tcl.evalf("set temp [$rlg lookup %d %d]",hop2,(int)hdr->endnode);
                nh = (char *)tcl.result();
                hop1=hop2;
                hop2=atoi(nh);
            }/*end of while loop*/
        }/*end of if POSITIVE*/

        /*update the list of Requests*/
	Request newrequest(hdr->sender,hdr->endnode,hdr->bndw_request,hdr->start_time,hdr->stop_time);
        if(hdr->answer==POSITIVE)
	{
	     newrequest.SetStatus(SATISFIED);
        }
	else
	{
             newrequest.SetStatus(REJECTED);
	}
	list_request->add_to_list(newrequest);

	 /*cout<<"Hello  in the answer with list_size is "<<list_request_buffer->GetSize_of_list() <<"\n";*/
	/**************************************************
	 IF the list_request_buffer is not empty
         ELSE call recv with argument the packet which holds
	 the first request of the list_request_buffer
	****************************************************/
	if(list_request_buffer->GetSize_of_list()!=0)
	{
	  Request demand=list_request_buffer->get_first_rqst();
	  /*cout<<"Hello  from  answer"<<"\n";*/
	  /*demand.ShowRequest();*/
	  /*cout<<"Hello  again from answer"<<"\n";*/


          Packet* pkt = allocpkt();
          hdr_ip* hdrip = hdr_ip::access(pkt);/* Access the IP header*/
          hdrip->DSCP_=1;
          hdr_bbe* hdr = hdr_bbe::access(pkt);/* Access the bbe header*/
          hdr->sender =demand.GetSender();  /*address of the node(sender) who sends the request for bandwidth */
          hdr->endnode =demand.GetEndnode(); /*address of the last node of the request */
          hdr->hop_num=1;
          hdr->type=REQUEST;  /*packet type==REQUEST*/
          hdr->bndw_request=demand.GetBndw_request();
          hdr->start_time=demand.GetStart_time();/*when the allocation of the bandwith begins */
          hdr->stop_time=demand.GetStop_time();/*when the allocation of the bandwith ends */

          available=TRUE;/*the BBbase is now availiable*/

          recv(pkt, 0) ;
	}
	else
	{available=TRUE;/*the BBbase is now availiable*/}



    }/*end of packet handling, packet_type==ANSWER*/

    Packet::free(pkt);
    
}/*end of function recv*/


