/***************************************************************************
  Research Academic Computer Technology Institute,
  61 Riga Feraiou Str,26221 Patras ,Greece &
  Department of Computer Engineering and Informatics,
  University of Patras, 26500 Rion ,Patras,Greece

  Research Unit 6

  project name:Bandwidth Broker
  project team:Christos Bouras
               Dimitris Primpas
	       Kostas Stamos
	       Ioannis Pappas

  date:20 January 2005
  updated:no
  contact:bouras@cti.gr,primpas@cti.gr,stamos@cti.gr,pappasj@ceid.upatras.gr

*****************************************************************************/

/*Headers */
#include "BBedge.h"
#include <iostream.h>

/*constants*/
/*******************************************
 constants name:REQUEST,ANSWER,FIX_BWD
 		UNAVAILIABLE,DISTR_QUESTION,
		DISTR_ANSWER,CENTR_QUESTION,
		CENTR_ANSWER
mission:these constants specify the
	type of a packet
*******************************************/
#define REQUEST        0/*request packet from BBedge to BBbase for bandwidth */
#define ANSWER         1/*answer packet */
#define FIX_BWD        2/*command packet to BBedge to fix his bwd*/
#define UNAVAILABLE    3/*information packet to BBedge that BBbase is unavailiable now*/
#define DISTR_QUESTION 4 /*question packet from DistributedBBbase to BBedge if edge has bwd*/
#define DISTR_ANSWER   5 /*answer to  DistributedBBbase if BBedge has bwd*/
#define CENTR_QUESTION 6 /* question packet from CentralBBbase of what  BBedge controls */
#define CENTR_ANSWER   7 /*answer to CentralBBbase of what  BBedge controls*/


#define NEGATIVE 0
#define POSITIVE 1

/**********************************************
 constants name:DOWN,ACTIVE
 mission:these constants specify the status in
         which a BBagent is
**********************************************/
#define DOWN     0
#define ACTIVE   1


int hdr_bbb::offset_;

static class BBedgeHeaderClass : public PacketHeaderClass {
public:
	BBedgeHeaderClass() : PacketHeaderClass("PacketHeader/BBedge",
					      sizeof(hdr_bbb)) {
		bind_offset(&hdr_bbb::offset_);
	}
} class_bbbhdr;

static class BBedgeClass : public TclClass {
public:
	BBedgeClass() : TclClass("Agent/BBedge") {}
	TclObject* create(int, const char*const*) {
		return (new BBedgeAgent());
	}
} class_BBedge;


/*constructor*/
BBedgeAgent::BBedgeAgent() : Agent(PT_BBB)
{
   bind("packetSize_", &size_);
}


/************************************************
 function name: BBedgeAgent::command
 return value:int
 arguments:number of tcl command arguments,
 	   the command arguments
 mission:takes the tcl commands and transforms
         them to c++ functions & operations
*************************************************/
int BBedgeAgent::command(int argc, const char*const* argv)
{
    if (argc == 4)/*if we have four(4) arguments*/
    {
        if (strcmp(argv[1], "sendto") == 0)
        {
            return (sendto((nsaddr_t)atoi(argv[2]),atoi(argv[3])));
        }
    }

    if (argc == 6)/*if we have six(6) arguments*/
    {
       if (strcmp(argv[1], "sendto") == 0)
        {
            return (sendto((nsaddr_t)atoi(argv[2]),atoi(argv[3]),atof(argv[4]),atof(argv[5])));
        }
	/****************************************
	 if the tcl command is set_bndw ,then for
	 the connection with a specific
	 neighbour(argv[2]) of the BBedgeAgent,we
	 set a bandwidth argv[3] to be controlled
	 by the BBbase Agent.This information is
	 added to a list which we will ask,if we
	 want to reserve a specific amount of
	 bandwidth in the future from the BBedge
	 Agent to his neighbour.
	*****************************************/
        else if (strcmp(argv[1], "set_bndw") == 0)
        {
            neighbour neighb(atoi(argv[2]),atoi(argv[3]),atoi(argv[4]),atoi(argv[5]));
            nghbr_list.push_back(neighb);
            return (TCL_OK);
        }

    }

    if(argc ==4)/*if we have four(4) arguments*/
    {
      if (strcmp(argv[1], "inform") == 0)
        {
            return (inform(atoi(argv[2]),atoi(argv[3])));
        }
    }

    return (Agent::command(argc, argv));

}



/*******************************************************
 function name: BBedgeAgent::sendto
 return value:int
 arguments:a node address,bandwidth
 mission:sends a request packet from BBedge Agent to
         BBbase Agent for bandwidth from here to dstid
********************************************************/
int BBedgeAgent::sendto(nsaddr_t dstid, int bwd)
{
    /****************************
     allocate & send a new packet
    *****************************/
    Packet* pkt = allocpkt();
    hdr_ip* hdrip = hdr_ip::access(pkt); /* Access the IP header*/
    hdrip->DSCP_=1;
    hdr_bbe* hdr = hdr_bbe::access(pkt); /* Access the bbe header*/
    hdr->sender = here_.addr_; /*address of the node(sender) who sends the request for bandwidth */
    hdr->endnode = dstid; /*address of the last node of the request */
    hdr->hop_num=1;
    hdr->type=REQUEST;    /*packet type==REQUEST*/
    hdr->bndw_request=bwd;
    send(pkt, 0);

    return (TCL_OK);
}


/*******************************************************
 function name: BBedgeAgent::sendto
 return value:int
 arguments:a node address,bandwidth
 mission:sends a request packet from BBedge Agent to
         BBbase Agent for bandwidth from here to dstid
         from start(time) till stop(time)
********************************************************/
int BBedgeAgent::sendto(nsaddr_t dstid, int bwd,double start,double stop)
{
    /*****************************
     allocate & send a new packet
    ******************************/
    Packet* pkt = allocpkt();
    hdr_ip* hdrip = hdr_ip::access(pkt);/* Access the IP header*/
    hdrip->DSCP_=1;
    hdr_bbe* hdr = hdr_bbe::access(pkt);/* Access the bbe header*/
    hdr->sender = here_.addr_; /*address of the node(sender) who sends the request for bandwidth */
    hdr->endnode = dstid; /*address of the last node of the request */
    hdr->hop_num=1;
    hdr->type=REQUEST;  /*packet type==REQUEST*/
    hdr->bndw_request=bwd;
    hdr->start_time=start;/*when the allocation of the bandwith begins */
    hdr->stop_time=stop;/*when the allocation of the bandwith ends */
    send(pkt, 0);

    return (TCL_OK);
}



/**********************************************
 function name: BBedgeAgent::inform
 return value:int
 arguments:start time and stop time
 mission:informs the CentralBBbase for his
         current condition
***********************************************/
int BBedgeAgent::inform(int start_time,int stop_time)
{
    /*****************************
     allocate & send a new packet
    ******************************/
    Packet* pkt = allocpkt();
    hdr_ip* hdrip = hdr_ip::access(pkt);/* Access the IP header*/
    hdrip->DSCP_=1;
    hdr_bbe* hdr = hdr_bbe::access(pkt);/* Access the bbe header*/
    hdr->type=CENTR_ANSWER; /*packet type==CENTR_ANSWER*/
    hdr->status=ACTIVE;     /*current status of the BBedge is ACTIVE*/

    /************************************************
     now i create an unsigned char array where i
     will enter the data of BBedge's list.
     (list<neighbour> nghbr_list)
     Then i pass these data into the packet which i
     will send to the CentralBBbase Agent
    *************************************************/

    int* data;/*array where i will enter the data of list*/

    data = new int[(int)(2*(nghbr_list.size()))]; /*dynamic assigned of the array*/

    /*cout<<"the magnitude of the BBedge's list is  "<< nghbr_list.size() <<"\n";*/


    int j=0; /*variable in order to move into the array*/
    typedef list<neighbour>::const_iterator LI; /* an iterator for nghbr_list */
    for(LI i=nghbr_list.begin();i!=nghbr_list.end();i++)
    {
        data[j] =i->n_id;     /*we enter the data like this:n_id,bwdth,n_id,bwdth...*/
	                      /*cout<<"the value of data[j] with node id  is  "<<data[j]<<"\n";*/
        j=j+1;                /*into the array*/
        data[j] =i->bwdth[0];
	                      /*cout<<"the value of data[j] with bandwidth is  "<<data[j]<<"\n"; */
	                      /*cout<<"the value of i->bwdth[0] is "<<i->bwdth[0]<<"\n"; */
	j=j+1;
    }
	                      /*cout<<"Hello after for-loop at BBedge::inform"<<"\n";*/

    PacketData *pData=new PacketData((int)(2*(nghbr_list.size())),start_time,stop_time);/*object of the class PacketData*/
	                                   /*cout<<"Hello after pData object creation at BBedge::inform"<<"\n";*/
    pData->SetNeighData_(data); /* i pass the data into the object*/
                                    /*cout<<"Hello after pData->SetNeighData_(data) function at BBedge::inform"<<"\n";*/

    pkt->setneighdata(pData); /*i pass the data into the packet*/
                                   /*cout<<"Hello after pkt->setneighdata(pData)  function at BBedge::inform"<<"\n";*/

    send(pkt, 0);
	                      /*cout<<"Hello after send  function at BBedge::inform"<<"\n";*/

    return (TCL_OK);
}



/***********************************************
 function name: BBedgeAgent::recv
 return value:void
 arguments:a packet , a handler
 mission:receives and handles the packets of
         the BBedgeAgent
************************************************/
void BBedgeAgent::recv(Packet* pkt, Handler*)
{
    /* Access the IP header for the received packet*/
    hdr_ip* hdrip = hdr_ip::access(pkt);
    /* Access the bbb header for the received packet*/
    hdr_bbb* hdr = hdr_bbb::access(pkt);


    /*****************************************
     if the type of the packet is REQUEST
    ******************************************/
    if (hdr->type == REQUEST)
    {
        /***********************************************
	 IF i am the last node of the request,it means
	 that all the previous nodes answered positive.
	 So i send to the BBbase Agent a POSITIVE
	 answer packet.
        ************************************************/
        if (here_.addr_ == hdr->endnode )
        {
	    /*****************************
             allocate & send a new packet
            ******************************/
            Packet* next_pkt = allocpkt();
            hdr_ip* ippkt = hdr_ip::access(next_pkt);/* Access the IP header*/
            ippkt->DSCP_=1;
            hdr_bbe* next_hdr = hdr_bbe::access(next_pkt);/* Access the bbe header*/
            next_hdr->sender = hdr->sender;
            next_hdr->endnode = hdr->endnode;
            next_hdr->bndw_request = hdr->bndw_request;
            next_hdr->type = ANSWER; /* packet type==ANSWER*/
            next_hdr->answer=POSITIVE;
	    next_hdr->start_time = hdr->start_time;
	    next_hdr->stop_time = hdr->stop_time;
            send(next_pkt,0);
        }
        /************************************************
	 ELSE send back to the BBbase a packet.
	 At start set as packet's type the value REQUEST.
	 Then check the availiability of the bandwidth
	 that it has been requested for the connection
	 with a specific bandwidth.If the answer is
	 positive do nothing and send the packet as is
	 in order to make clear to the BBbase that this
	 BBedge answered positive.Else if the answer is
	 negative change the packet's type with the
	 value ANSWER and set the packet's answer as
	 NEGATIVE.
	*************************************************/
        else
        {
	    /*****************************
             allocate & send a new packet
            ******************************/
            Packet* next_pkt = allocpkt();
            hdr_ip* ippkt = hdr_ip::access(next_pkt);/* Access the IP header*/
            ippkt->DSCP_=1;
            hdr_bbe* next_hdr = hdr_bbe::access(next_pkt);/* Access the bbe header*/
            next_hdr->sender = hdr->sender;
            next_hdr->endnode = hdr->endnode;
            next_hdr->type = REQUEST; /*packet's type== REQUEST*/
            next_hdr->bndw_request = hdr->bndw_request;
	    next_hdr->start_time = hdr->start_time;
	    next_hdr->stop_time = hdr->stop_time;
            next_hdr->hop_num = 1+hdr->hop_num;
            /*finding his neighbour*/
            Tcl& tcl= Tcl::instance();
            tcl.evalc("$ns compute-routes");
            tcl.evalc("set rlg [$ns get-routelogic]");
            tcl.evalf("set temp [$rlg lookup %d %d]",(int)here_.addr_,(int)hdr->endnode);
            char *nh = (char *)tcl.result();
            int avail_bndw=0;

	    /*check the availiable bandwidth from me till nh node */
            typedef list<neighbour>::const_iterator LI;
            for(LI i=nghbr_list.begin();i!=nghbr_list.end();i++)
            {
                if (i->n_id==atoi(nh))
                {
		  int j;
		  for(j=hdr->start_time;j<hdr->stop_time;j++)
		  {
		     if(i->bwdth[j]<hdr->bndw_request)/*if the availiable bandwidth is less than the requested*/
		     {
	                next_hdr->type = ANSWER;
                        next_hdr->answer = NEGATIVE;
                        break;
		     }
		  }
		 break;
		}
            }

	   send(next_pkt,0);
        }
    } /* end of  "if (hdr->type == REQUEST)" */


    /*****************************************
     if the type of the packet is ANSWER
    ******************************************/
    /******************************************
     the BBedge Agent receives an answer about
     his request for bandwidth
    ******************************************/
    else if (hdr->type==ANSWER)
    {
        if (hdr->answer==POSITIVE)
            cout <<"from     :"<<here_.addr_<<" to    :"<<(int)hdr->endnode<<" for the time  "<<hdr->start_time<<"-"<<hdr->stop_time <<" The answer is YES\n";
        else
            cout <<"from     :"<<here_.addr_<<" to    :"<<(int)hdr->endnode<<" for the time  "<<hdr->start_time<<"-"<<hdr->stop_time <<" The answer is NO\n";
        char out[10000];

        sprintf(out, "%s recv %d %d %d %d %d %d", name(),hdr->endnode,hdr->answer,hdr->dscp,hdr->bndw_request,(int)hdr->start_time,(int)hdr->stop_time);
        Tcl& tcl = Tcl::instance();
        tcl.eval(out);
    }/* end of "else if (hdr->type==ANSWER)" */


    /*****************************************
     if the type of the packet is FIX_BWD
    ******************************************/
    /*****************************************
     the BBedge Agent receives a command
     packet to reduce his bandwidth on the
     connection with his neighbour
     hdr->endnode(n_id) like this:
     bandwidth=bandwidth - hdr->bndw_request
    ******************************************/
    else if (hdr->type==FIX_BWD)
    {
        typedef list<neighbour>::iterator LI;
        for(LI i=nghbr_list.begin();i!=nghbr_list.end();i++)
        {
            if (i->n_id==hdr->endnode)
	    {
               int j;
		  for(j=hdr->start_time;j<hdr->stop_time;j++)
		  {
		     i->bwdth[j]=i->bwdth[j] - hdr->bndw_request;
		  }
		 break;
	    }
        }
    }/* end of "else if (hdr->type==FIX_BWD)" */


    /******************************************
     if the type of the packet is UNAVAILIABLE
    *******************************************/
    /******************************************
     the BBedge Agent receives an answer that
     his request for bandwidth cannot be
     done cause the BBbase is not
     availiable
    ******************************************/
    else if (hdr->type==UNAVAILABLE)
    {
        cout <<"from     :"<<here_.addr_<<" to    :"<<(int)hdr->endnode<<" BB_BASE NOT AVAILABLE\n";
    }/*end of "else if (hdr->type==UNAVAILABLE)" */


    /********************************************
     if the type of the packet is DISTR_QUESTION
    *********************************************/
    else if (hdr->type==DISTR_QUESTION)
    {
        /***********************************************
	 IF i am the last node of the request, i send
	 to the DistributedBBbase Agent a POSITIVE
	 answer packet.
        ************************************************/
        if (here_.addr_ == hdr->endnode )
        {
	    /*****************************
             allocate & send a new packet
            ******************************/
            Packet* next_pkt = allocpkt();
            hdr_ip* ippkt = hdr_ip::access(next_pkt);/* Access the IP header*/
            ippkt->DSCP_=1;
            hdr_bbe* next_hdr = hdr_bbe::access(next_pkt);/* Access the bbe header*/
            next_hdr->sender = hdr->sender;
            next_hdr->endnode = hdr->endnode;
            next_hdr->bndw_request = hdr->bndw_request;
            next_hdr->type = DISTR_ANSWER;/* packet type==DISTR_ANSWER*/
            next_hdr->answer=POSITIVE;
	    next_hdr->start_time=hdr->start_time;
	    next_hdr->stop_time=hdr->stop_time;
            send(next_pkt,0);
        }
        /************************************************
	 ELSE i am not the last node of the request and
	 so i send a POSITIVE or a NEGATIVE answer packet
	 to the DistributedBBbase Agent
	*************************************************/
        else
        {
	    /*****************************
             allocate & send a new packet
            ******************************/
            Packet* next_pkt = allocpkt();
            hdr_ip* ippkt = hdr_ip::access(next_pkt);/* Access the IP header*/
            ippkt->DSCP_=1;
            hdr_bbe* next_hdr = hdr_bbe::access(next_pkt);/* Access the bbe header*/
            next_hdr->sender = hdr->sender;
            next_hdr->endnode = hdr->endnode;
            next_hdr->type = DISTR_ANSWER;/* packet type==DISTR_ANSWER*/
            next_hdr->bndw_request = hdr->bndw_request;
	    next_hdr->start_time=hdr->start_time;
	    next_hdr->stop_time=hdr->stop_time;
            next_hdr->hop_num = 1+hdr->hop_num;
	    next_hdr->answer = POSITIVE;
            /*finding his neighbour*/
            Tcl& tcl= Tcl::instance();
            tcl.evalc("$ns compute-routes");
            tcl.evalc("set rlg [$ns get-routelogic]");
            tcl.evalf("set temp [$rlg lookup %d %d]",(int)here_.addr_,(int)hdr->endnode);
            char *nh = (char *)tcl.result();
            int avail_bndw=0;

	    /*check the availiable bandwidth from me till nh node */
            typedef list<neighbour>::const_iterator LI;
            for(LI i=nghbr_list.begin();i!=nghbr_list.end();i++)
            {
                if (i->n_id==atoi(nh))
                {
		  int j;
		  for(j=hdr->start_time;j<hdr->stop_time;j++)
		  {
		     if(i->bwdth[j]<hdr->bndw_request)/*if the availiable bandwidth is less than the requested*/
		     {
	                next_hdr->answer = NEGATIVE;/*NEGATIVE answer to DistributedBBbase*/
                     }
		  }
		 break;
		}
            }

            send(next_pkt,0);
        }

    }/*end of "else if (hdr->type==DISTR_QUESTION)" */



    /********************************************
     if the type of the packet is CENTR_QUESTION
	FUTURE USE
    *********************************************/
    //    else if (hdr->type==CENTR_QUESTION)
    //  {
     //  inform();
      //}/*end of "else if (hdr->type==CENTR_QUESTION)" */

    Packet::free(pkt);
}/*end of funtion recv*/


